
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: rostami@google.com (Afshin Rostamizadeh)

#ifndef KERNEL_LIB_KRR_H_
#define KERNEL_LIB_KRR_H_

#include "kernel/lib/dense-matrix.h"
#include "kernel/lib/dense-vector.h"

namespace kernel {

//
// Forward Declarations
//

//TODO(rostami): Instead of precomputing the kernels, pass only the
//kernel function and data.  Then the kernel computations can be done
//on-the-fly at the last moment. Also, the user then does not need to
//calculate the kernel matrix outside of the call to KRRxxx.

//
// Notation:
// m = number of training points
// n = number of test points
// N = dimensionality of feature space
//

// Compute the optimal dual variables to the kernel ridge regression
// (KRR) optimization problem:
//
//   max_alpha -lambda ||alpha||^2 - alpha^T K alpha + 2 alpha^T y
//
//   where:
//   K = kernel matrix (m-by-m)
//   y = labels
//   lambda = reg_param
//   alpha = dual_vars
//
//   Completxity: Solves a system of equations with an order m constraint
//   matrix.
template <typename T>
void KRRDualTrain(const DenseMatrix<T> &kernel, const DenseVector<T> &labels,
              double reg_param, DenseVector<T> *dual_vars);

// Output KRR prediction given the n-by-m test kernel (n test points,
// m training points) and optimal dual variables from KRRDualTrain.
template <typename T>
void KRRDualPredict(const DenseMatrix<T> &test_kernel,
                const DenseVector<T> &dual_vars,
                DenseVector<T> *sol);

//Solve the KRR problem in the primal:
//
//    min_w lambda ||w||^2 + ||X w - y||^2
//
//    where:
//    X = explicit feature mapping (m-by-N)
//    y = labels
//    lambda = reg_param
//    w = primal_vars
//
//   Completxity: Solves a system of equations with an order N constraint
//   matrix.
template <typename T, class M>
void KRRPrimalTrain(const M &feature_map, const DenseVector<T>
                    &labels, double reg_param, DenseVector<T> *primal_vars);


// Output KRR prediction given the n-by-N feature mapping of test
// points and optimal primal variables from KRRPrimalTrain.
template <typename T, class M>
void KRRPrimalPredict(const M &test_feature_map,
                const DenseVector<T> &primal_vars,
                DenseVector<T> *sol);

// Compute the optimal dual variables to the ridge regression problem
// (KRR) using an approximate SVD.  Suitable for cases when the entire
// kernel matrix is too expensive (or impossible) to compute.
//
// INPUT:
// feature_map - a Dense or SemiDense matrix with each row corresponding
//   to a training instance.
// sample_index - the number of columns to be sampled from the kernel matrix,
//   thus the size of the largest matrix computed is
//   feature_map.NumRow() * sample_index.Size(), as oppose to
//   feature_map.NumRow()^2.
// labels - the training labels for instance.
// reg_param - the regularization parameter for the KRR problem.
// OUTPUT:
// dual_vars - the approximate dual variable solution to the KRR problem.
template <typename T, class M>
void KRRApproxDualTrain(const M &feature_map,
                        const DenseVector<size_t> &sample_index,
                        const DenseVector<T> &labels,
                        double reg_param,
                        DenseVector<T> *dual_vars);

//
// Implementations
//

template <typename T>
void KRRDualTrain(const DenseMatrix<T> &kernel, const DenseVector<T> &labels,
              double reg_param, DenseVector<T> *dual_vars) {
  CHECK(kernel.NumRow() == kernel.NumColumn());

  DenseMatrix<T> U(kernel);
  //U = K + reg_param * I
  for(size_t i = 0; i < U.NumRow(); ++i)
    U.SetElement(i, i, U.Element(i, i) + reg_param);

  DenseMatrix<T> b(labels);
  DenseMatrix<T> x(U.NumColumn(), 1, 0);

  //U * x = b
  SolveLinearEq(U, b, &x);

  dual_vars->Resize(b.NumRow(), 0);
  for(size_t i = 0; i < b.NumRow(); ++i)
    dual_vars->SetElement(i, x.Element(i,0));
}

template <typename T>
void KRRDualPredict(const DenseMatrix<T> &test_kernel,
                const DenseVector<T> &dual_vars,
                DenseVector<T> *sol) {
  //sol = Ktst * alpha
  //sol(x) = sum_i alpha_i phi(x_i) * x
  Product(test_kernel, dual_vars, sol);
}

template <typename T, class M>
void KRRPrimalTrain(const M &feature_map, const DenseVector<T>
                    &labels, double reg_param, DenseVector<T> *primal_vars) {
  CHECK(feature_map.NumRow() == labels.Size());

  DenseMatrix<T> U(feature_map.NumColumn(), feature_map.NumColumn(), 0);
  //U = X^T X + reg_param * I
  typename M::TransposeMatrix feature_map_t;
  feature_map.Transpose(&feature_map_t);
  Product(feature_map_t, feature_map, &U);
  for(size_t i = 0; i < U.NumRow(); ++i)
    U.SetElement(i, i, U.Element(i, i) + reg_param);

  DenseVector<T> temp_v(feature_map.NumColumn());
  Product(feature_map_t, labels, &temp_v);
  DenseMatrix<T> b(temp_v);
  DenseMatrix<T> x(U.NumRow(), 1, 0);

  //U * x = b
  SolveLinearEqPDS(U, b, &x);

  primal_vars->Resize(x.NumRow(), 0);
  for(size_t i = 0; i < x.NumRow(); ++i)
    primal_vars->SetElement(i, x.Element(i, 0));
}

template <typename T, class M>
void KRRPrimalPredict(const M &test_feature_map,
                const DenseVector<T> &primal_vars,
                DenseVector<T> *sol) {
  //sol(x) = phi(x) * w
  Product(test_feature_map, primal_vars, sol);
}

//
//  Approximate methods
//

template <typename T, class M>
void KRRApproxDualTrain(const M &feature_map,
                        const DenseVector<size_t> &sample_index,
                        const DenseVector<T> &labels,
                        double reg_param,
                        DenseVector<T> *dual_vars) {
  CHECK(feature_map.NumRow() == labels.Size());
  M *feature_map_sample = new M();
  typename M::TransposeMatrix *feature_map_sample_t
      = new typename M::TransposeMatrix();
  SubMatrixCopy(feature_map, sample_index, feature_map_sample);
  feature_map_sample->Transpose(feature_map_sample_t);
  delete feature_map_sample;

  DenseMatrix<T> Evec;
  DenseMatrix<T> *C = new DenseMatrix<T>;
  DenseVector<T> Eval;
  Product(feature_map, *feature_map_sample_t, C);
  delete feature_map_sample_t;

  ApproxSymEigen(sample_index, C, &Eval);
  //rename
  Evec = *C;
  delete C;

  // (lambda I + K)^-1 y
  //    = 1/lambda * (y - E (lambda I + S E^T E)^{-1} S E^T y)
  DenseVector<T> temp_v, temp_u;
  DenseMatrix<T> Evec_t, R;
  DenseMatrix<T> *S = new DenseMatrix<T>();
  DenseMatrix<T> *temp_m = new DenseMatrix<T>();
  Evec.Transpose(&Evec_t);
  Diagonal(Eval, S);


  // R = S E^T
  Product(*S, Evec_t, &R);
  //free memory
  delete S;

  // temp_m = lambda I + S E^T E
  Product(R, Evec, temp_m);
  CHECK(temp_m->NumRow() == temp_m->NumColumn());
  for (size_t i = 0; i < temp_m->NumRow(); ++i)
    temp_m->SetElement(i, i, temp_m->Element(i, i) + reg_param);

  DenseMatrix<T> temp_m_inv;

  Inverse(*temp_m, &temp_m_inv);
  //free memory
  delete temp_m;

  // temp_v = R y = S E^T y
  Product(R, labels, &temp_v);
  // temp_u = (lambda I + S E^T E)^{-1} * temp_v
  Product(temp_m_inv, temp_v, &temp_u);
  // temp_v = E * temp_u
  Product(Evec, temp_u, &temp_v);

  // dual_vars =  1/lambda (y - temp_v)
  dual_vars->Resize(temp_v.Size(), 0);
  for (size_t i = 0; i < temp_v.Size(); i++)
    dual_vars->SetElement(i, 1 / reg_param * (labels.Element(i) - temp_v.Element(i)));
}

template <typename T, class M>
void KRRApproxPrimalTrain(const M &feature_map, const DenseVector<T>
                    &labels, double reg_param, DenseVector<T> *primal_vars) {
  CHECK(feature_map.NumRow() == labels.Size());
  LOG(INFO) << "ERROR: KRRApproxPrimalTrain not yet implemented" << endl;
}

} //namespace kernel

#endif  // KERNEL_LIB_KRR_H_
