// sparse-vector.h
//

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// Author: allauzen@google.com (Cyril Allauzen)
//
// \file
// A simple reference-counted sparse vector representation

#ifndef KERNEL_LIB_SPARSE_VECTOR_H__
#define KERNEL_LIB_SPARSE_VECTOR_H__

#include <vector>
#include <algorithm>

#include "kernel/lib/kernel.h" // change to min req


namespace kernel {

// Forward Declarations
template <typename T>
class DenseMatrix;

//
// SPARSE VECTOR CLASS
//

// Forward declarations
template <class T, class I>
class SparseVectorData;
template <class T>
class DenseVector;


// SparseVector: a simple reference-counted dense vector representation
template <class T, class I = uint32, class D = SparseVectorData<T, I> >
class SparseVector {
 public:
  typedef T ElementType;
  typedef I IndexType;
  typedef D DataType;

  SparseVector() : data_(new DataType) {}

  SparseVector(const SparseVector<T, I> &vec) : data_(vec.data_) {
    if (data_)
      data_->IncrRefCount();
  }

  ~SparseVector() {
    if (data_ && !data_->DecrRefCount())
      delete data_;
  }

  SparseVector<T, I> &operator=(const SparseVector<T, I> &v) {
    if (this != &v) {
      if (!data_->DecrRefCount()) delete data_;
      v.data_->IncrRefCount();
      data_ = v.data_;
    }
    return *this;
  }

  size_t Size() const { return data_->Size(); }

  size_t MaxIndex() const {
    if(data_ && data_->Size() > 0)
      return data_->Pair(data_->Size() - 1).first;
    else
      return 0;
  }

  void Clear() {
    MutateCheck();
    data_->Clear();
  }

  void Reserve(size_t n) { data_->Reserve(n); }

  const pair<I, T> &Pair(size_t i) const { return data_->Pair(i); }

  T Element(I i) const { return data_->Element(i); }

  void AddElement(I i, const T &x) {
    if (x == 0.0F)
      return;
    MutateCheck();
    data_->AddElement(i, x);
  }

  void SetElement(I i, const T &x) {
    MutateCheck();
    data_->SetElement(i, x);
  }

  void SetPair(size_t i, const pair<I, T> &p) {
    MutateCheck();
    data_->SetPair(i, p);
  }

  static SparseVector<T, I> *Read(istream &strm) {
    return new SparseVector<T, I>(SparseVectorData<T, I>::Read(strm));
  }

  bool Write(ostream &strm) const {
    return data_->Write(strm);
  }

 private:
  SparseVector(DataType *data) : data_(data) {}

  void MutateCheck() {
    // Copy on write
    if (data_->RefCount() > 1) {
      data_->DecrRefCount();
      data_ = new DataType(*data_);
    }
  }

  DataType *data_;
};


template <typename T, typename I>
class SparseVectorData {
 public:
  SparseVectorData() {}

  explicit SparseVectorData(const SparseVectorData<T, I> &data)
      : pairs_(data.pairs_) {}

  ~SparseVectorData() {}

  size_t Size() const { return pairs_.size(); }
  void Reserve(size_t n) { pairs_.reserve(n); }
  void Clear() { pairs_.clear(); }

  const pair<I, T> &Pair(size_t i) const { return pairs_[i]; }

  T Element(I i) const {
    typename vector<pair<I, T> >::const_iterator iter = lower_bound(
        pairs_.begin(), pairs_.end(), make_pair(i, 0.0F), Compare());
    return (iter != pairs_.end()) && ((*iter).first == i)
        ? (*iter).second : 0.0F;
  }

  void AddElement(I i, const T &x) {
    if (!pairs_.empty() && (pairs_.back().first >= i))
        LOG(FATAL) << "SparseVector::AddElement: non-zero elements must"
                   << " be added in order";
    pairs_.push_back(make_pair(i, x));
  }

  void SetElement(I i, const T &x) {
    typename vector<pair<I, T> >::iterator iter = lower_bound(
        pairs_.begin(), pairs_.end(), make_pair(i, 0.0F), Compare());
    if ((*iter).first == i) {
      if (x == 0.0F)
        pairs_.erase(iter);
      else
        (*iter).second = x;
    } else {
      pairs_.insert(iter, make_pair(i, x));
    }
  }

  void SetPair(size_t i, const pair<I, T> &p) {
    pairs_[i] = p;
  }

  int RefCount() const { return ref_count_.count(); }
  int IncrRefCount() { return ref_count_.Incr(); }
  int DecrRefCount() { return ref_count_.Decr(); }

  static SparseVectorData<T, I> *Read(istream &strm) {
    SparseVectorData<T, I> *data = new SparseVectorData<T, I>;
    fst::ReadType(strm, &(data->pairs_));
    if (!strm) {
      LOG(ERROR) << "DenseVector::Read: Read failed!";
      return 0;
    }
    return data;
  }

  bool Write(ostream &strm) const {
    fst::WriteType(strm, pairs_);
    if (!strm) {
      LOG(ERROR) << "DenseVector::Write: write failed!";
      return false;
    }
    return true;
  }

 private:

  struct Compare {
    bool operator()(const pair<I, T> &x, const pair<I, T> &y) {
      return x.first < y.first;
    }
  };

  vector<pair<I, T> > pairs_;
  RefCounter ref_count_;

  void operator=(const SparseVectorData<T, I> &);  // disallow
};


//
// VECTOR OPERATIONS
//

// Vector equality
template <typename V1, typename V2> inline
bool Equal(const V1 &a, const V2 &b) {
  if (a.Size() != b.Size())
    return false;
  for (size_t i = 0; i < a.Size(); ++i)
    if (a.Pair(i) != b.Pair(i))
      return false;
  return true;
}

// Vector approximate equality
template <typename V> inline
bool ApproxEqual(const V &a, const V &b, typename V::ElementType delta) {
  if (a.Size() != b.Size())
    return false;
  for (size_t i = 0; i < a.Size(); ++i) {
    if (a.Pair(i).first != b.Pair(i).first)
      return false;
    if ((a.Pair(i).second > b.Pair(i).second + delta) ||
        (a.Pair(i).second < b.Pair(i).second - delta))
      return false;
  }
  return true;
}

//
// Multiplication by a scalar
//

// y = alpha * x
template <typename T, typename I, typename V> inline
void ScalarProduct(const T &a,
                   const V &x,
                   SparseVector<T, I> *y) {
  y->Clear();
  for (size_t i = 0; i < x.Size(); ++i)
    y->AddElement(x.Pair(i).first, x.Pair(i).second * a);
}

// x = alpha * x
template <typename T, typename I> inline
void ScalarProduct(const T &alpha, SparseVector<T, I> *x) {
  for(size_t i = 0; i < x->Size(); ++i)
    x->SetPair(i, make_pair(x.Pair(i).first,
                            alpha * x.Pair(i).second));
}


//
// Addition
//

// z = alpha * x + y
template <typename T, typename I, typename V1, typename V2>
inline void Sum(const T &alpha,
                const V1 &x,
                const V2 &y,
                SparseVector<T, I> *z) {
  z->Clear();
  for (size_t i = 0, j = 0; i < x.Size(); ++i) {
    while ((j < y.Size()) && (y.Pair(j).first < x.Pair(i).first)) {
      z->AddElement(y.Pair(j).first,  y.Pair(j).second);
      ++j;
    }
    if (j < y.Size() && (x.Pair(i).first == y.Pair(j).first)) {
      z->AddElement(x.Pair(i).first,
                    alpha *x.Pair(i).second + y.Pair(j).second);
      ++j;
    } else {
      z->AddElement(x.Pair(i).first, alpha * x.Pair(i).second);
    }
  }
}
// y = alpha * x + y
template <typename T, typename I, typename V> inline
void Sum(const T &alpha, const V &x, SparseVector<T, I> *y) {
  SparseVector<T, I> z;
  Sum(alpha, x, *y, &z);
  *y = z;
}

// z = x + y
template <typename T, typename I, typename V1, typename V2> inline
void Sum(const V1 &x, const V2 &y, SparseVector<T, I> *z) {
  Sum<T, I>(1.0F, x, y, z);
}

// y = x + y
template <typename T, typename I, typename V> inline
void Sum(const V &x, SparseVector<T, I> *y) {
  Sum<T, I>(1.0F, x, y);
}


//
// Pointwise and dot products
//

// Point-wise product between two sparse vectors:
// z = x .* y
template <typename T, typename I, typename V1, typename V2> inline
void PointwiseProduct(const V1 &x,
                      const V2 &y,
                      SparseVector<T, I> *z) {
  z->Clear();
  for (size_t i = 0, j = 0; (i < x.Size()) && (j < y.Size()); ++i) {
    while ((j < y.Size()) && (y.Pair(j).first < x.Pair(i).first))
      ++j;
    if (j == y.Size()) break;
    if (x.Pair(i).first == y.Pair(j).first) {
      z->AddElement(x.Pair(i).first, x.Pair(i).second * y.Pair(j).second);
      ++j;
    }
  }
}

// Point-wise product between one sparse and one dense vector:
// z = x .* y
template <typename T, typename I, typename V> inline
void PointwiseProduct(const V &x,
                      const DenseVector<T> &y,
                      SparseVector<T, I> *z) {
  z->Clear();
  for (size_t i = 0; i < x.Size(); ++i) {
    if (x.Pair(i).first >= y.Size()) // append zeros to y
      continue;
    z->AddElement(x.Pair(i).first,
                  x.Pair(i).second * y.Element(x.Pair(i).first));
  }
}

// Dot product between two sparse vectors:
// dot = x^T * y
template <typename V1, typename V2> inline
typename V1::ElementType DotProduct(const V1 &x, const V2 &y) {
  typename V1::ElementType dot = 0.0F;
  for (size_t i = 0, j = 0; (i < x.Size()) && (j < y.Size()); ++i) {
    while ((j < y.Size()) && (y.Pair(j).first < x.Pair(i).first))
      ++j;
    if (j == y.Size()) break;
    if (x.Pair(i).first == y.Pair(j).first) {
      dot += x.Pair(i).second * y.Pair(j).second;
      ++j;
    }
  }
  return dot;
}

// Dot product between one sparse and one dense vector:
// dot = x^T * y
template <typename T, typename V> inline
T DotProduct(const V &x, const DenseVector<T> &y) {
  T dot = 0.0F;
  for (size_t i = 0; (i < x.Size()) && (x.Pair(i).first < y.Size()); ++i)
    dot += x.Pair(i).second * y.Element(x.Pair(i).first);

  return dot;
}

// Dot product between two sparse vector:
// dot = x^T * y
// Uses binary search to find element in larger vector.
template <typename T, typename I> inline
T DotProductSkew(const SparseVector<T, I> &big,
                 const SparseVector<T, I> &small) {
  T dot = 0.0F;
  T val;
  for (size_t i = 0; i < small.Size(); ++i)
    if ((val = big.Element(small.Pair(i).first)) != 0)
      dot += val * small.Pair(i).second;

  return dot;
}

// Weighted dot product:
// dot = x^T * Diag(y) * z
template <typename T, typename V1, typename V2> inline
T DotProduct(const V1 &x,
             const DenseVector<T>  &y,
             const V2 &z) {
  T dot = 0.0F;
  for (size_t i = 0, j = 0; (i < x.Size()) && (j < z.Size()); ++i) {
    while ((j < z.Size()) && (z.Pair(j).first < x.Pair(i).first))
      ++j;
    if (j == z.Size()) break;
    if ((x.Pair(i).first == z.Pair(j).first) && (x.Pair(i).first < y.Size())) {
      dot += x.Pair(i).second * y.Element(x.Pair(i).first) * z.Pair(j).second;
      ++j;
    }
  }
  return dot;
}

// Weighted dot product:
// dot = x^T * Diag(y) * z
template <typename V> inline
typename V::ElementType DotProduct(const V &x,
             const V &y,
             const V &z) {
  SparseVector<typename V::ElementType, typename V::IndexType> t;
  PointwiseProduct(x, y, &t);
  return DotProduct(t, z);
}

template <typename T, typename I> inline
T OuterProduct(const SparseVector<T, I> &x,
         size_t size_x,
         const SparseVector<T, I> &y,
         size_t size_y, 
         DenseMatrix<T> *mat) {
  mat->Reset(size_x, size_y, 0);
  for (size_t i = 0; i < x.Size(); ++i)
    for (size_t j = 0; j < y.Size(); ++j)
      mat->SetElement(x.Pair(i).first, y.Pair(j).first,
                     x.Pair(i).second * y.Pair(j).second);
}  

}  // namespace kernel

#endif  // KERNEL_LIB_SPARSE_VECTOR_H__
